/***************************************************************************
 *                           farray.h  - a simple implementation of
 *                                       a scalable array that knows
 *                                       its own size.
 *                                 -------------------
 *   begin                : 2003
 *   copyright            : (C) 2003 by Siegfried Nijssen
 *   email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef FARRAY_H
#define FARRAY_H
#include <stdlib.h>
#include <string.h>
#define INITSIZE 20
#define INCREASE 15
#include "type.h"

template<class T>
class FArray {
  private:
    int _size;  // real size;
    int filled; // zerod size;
  public:
    T *array; // this one is public!!! The sole reason is to make access FAST!
    int size () const { return filled; }
    FArray<T> () {
      _size = -1;
      filled = 0;
    }
    FArray<T> ( int initsize ) {
      array = (T*) malloc ( sizeof ( T ) * initsize );
      _size = initsize;
      filled = 0;
    }
    ~FArray<T> () {
      if ( _size != -1 )
        free ( array );
    }
    inline T &operator[] ( int index ) const {
      return array[index];
    }
    inline T element ( int index ) const {
      if ( index >= filled )
        return T ( 0 );
      else
        return array[index];
    }
    // allows for using the array as a random access stack:
    void append ( T &value ) {
      insertAtRef ( size (), value );
    }
    void releaselast () {
      filled--;
    }
    // if the array is not expected to grow any more, this function
    // can be used to free unused space.
    void compact () {
      if ( _size - filled > 2 ) { // some threshold
        array = (T*) realloc ( array, sizeof ( T ) * filled );
        _size = filled;
      }
    }
    void insertAtRef ( int position, T &value ) {
      if ( position < filled ) // optimised for insertions in allocated space.
        array[position] = value;
      else
	if ( position < _size ) {
	  if ( position == filled )
	    filled++;
	  else {
	    bzero ( array + filled, ( position - filled ) * sizeof ( T ) );
	    filled = position + 1;
	  }
          array[position] = value;
	}
	else if ( _size != -1 ) {
          int newsize = position + INCREASE;
          array = (T*) realloc ( array, sizeof ( T ) * newsize );
	  _size = newsize;
	  if ( position > filled ) {
	    bzero ( array + filled, ( position - filled ) * sizeof ( T ) );
	  }
	  filled = position + 1;
	  array[position] = value;
	}
	else {
          array = (T*) malloc ( sizeof ( T ) * ( position + INCREASE ) );
	  bzero ( array, position * sizeof ( T ) );
          _size = position + INCREASE;
	  array[position] = value;
          filled = position + 1;
	}
    }
    // sets the size to the given size. 
    // Unlike the insertAt function, this function leaves
    // the allocated space unzerod.
    void resize ( int size ) {
      if ( size <= _size ) // optimised for insertions in allocated space.
        filled = size;
      else
        if ( _size != -1 ) {
          int newsize = size + INCREASE;
          array = (T*) realloc ( array, sizeof ( T ) * newsize );
	  _size = newsize;
	  filled = size;
	}
	else {
          array = (T*) malloc ( sizeof ( T ) * ( size + INCREASE ) );
          _size = size + INCREASE;
          filled = size;
	}
    }
    inline void insertAt ( int position, T value ) {
      insertAtRef ( position, value );
    }
};

template<class T>
class FPArray {
  private:
    int _start, _end;
    int filledstart, filledend;
  public:
    T *array; // this one is public!!! The sole reason is to make access FAST!
    int size () const { return _start - _end + 1; }
    FPArray<T> () {
      _start = _end = -1000;
      filledstart = filledend = -1000;
    }
    FPArray<T> ( int initsize ) {
      array = (T*) malloc ( sizeof ( T ) * initsize );
      _start = 0;
      _end = initsize - 1;
      filledstart = filledend = -1000;
    }
    FPArray<T> ( int start, int end ) {
      _start = start;
      _end = end;
      array = (T*) malloc ( sizeof ( T ) * ( start - end + 1 ) );
      filledstart = filledend = -1000;
    }
    ~FPArray<T> () {
      if ( _start != -1 )
        free ( array );
    }
    inline T &operator[] ( int index ) const {
      return array[index-_start];
    }
    inline T element ( int index ) const {
      if ( index > filledend || index < filledstart )
        return T ( 0 );
      else
        return array[index-_start];
    }
    // if the array is not expected to grow any more, this function
    // can be used to free unused space.
    void compact () {
      if ( _end - _start - filledend + filledstart > 2 ) { // some threshold
        array = (T*) realloc ( array, sizeof ( T ) * ( filledend - filledstart + 1 ) );
	_start = filledstart;
	_end = filledend;
      }
    }
    void insertAtRef ( int position, T &value ) {
      if ( position <= filledend )
        if ( position >= filledstart )
	  array[position-_start] = value;
	else {
	  int diff = position - _start;
	  if ( diff < 0 ) {
	    int newsize = _end - position + INCREASE + 1;
	    T *newarray = (T*) malloc ( sizeof ( T ) * newsize );
	    memcpy ( newarray + filledstart - ( position - INCREASE ), array + filledstart - _start, sizeof ( T ) * ( filledend - filledstart + 1 ) );
	    free ( array );
	    array = newarray;
	    _start = position - INCREASE;
	    diff = INCREASE;
	  }
	  bzero ( array + diff + 1, ( filledstart - position - 1 ) * sizeof ( T ) );
	  array[diff] = value;
	  filledstart = position;
	}
      else {
        int diff = position - _start;
        if ( position >= _end ) {
          if (_start == -1000 ) {
	    array = (T*) malloc ( sizeof ( T ) * INITSIZE );
	    filledstart = filledend = position;
	    _start = position - INITSIZE / 2;
	    _end = position + INITSIZE / 2 - 1;
	    array[INITSIZE / 2] = value;
	    return;
	  }
	  else {
	    int newsize = diff + INCREASE + 1;
	    array = (T*) realloc ( array, sizeof ( T ) * newsize );
	    _end = position + INCREASE;
	  }
	}
	else
	  if ( filledstart == -1 ) {
	    array[position - _start] = value;
	    filledstart = filledend = position;
	    return;
	  }
	bzero ( array + filledend - _start + 1, ( position - filledend - 1 ) * sizeof ( T ) );
	array[diff] = value;
	filledend = position;
      }
    }
    inline void insertAt ( int position, T value ) {
      insertAtRef ( position, value );
    }
};


template<class T>
class FMultiArray {
  private:
    FPArray<void *> *data;
    int _dimensions;
    void recursive_delete ( FPArray<void*> *walk, int depth ) {
      for ( int i = 0; i < walk->size (); i++ )
        if ( walk->element ( i ) != 0 )
	  if ( depth )
	    recursive_delete ( (FPArray<void*>*) (*walk) [ i ], depth - 1 );
	  else
	    delete (T*) (*walk) [ i ];
      delete walk;
    }
  public:
    FMultiArray ( int dimensions ) {
      data = new FPArray<void*>;
      _dimensions = dimensions;
    }
    ~FMultiArray () {
      recursive_delete ( data, _dimensions - 1 );
    }
    void insertAt ( INTTYPE *position, T *value ) {
      int i, size = _dimensions - 1;
      FPArray<void *> *walk = data, *temp;
      for ( i = 0; i < size; i++ ) {
        if ( ( temp = (FPArray<void*>*) walk->element ( position [ i ] ) ) == 0 )
	  walk->insertAt ( position[i], (void*) ( temp = new FPArray<void*> ) );
	walk = temp;
      }
      walk->insertAt ( position[i], (void*) value );
    }
    inline T* operator[] ( INTTYPE *position ) const {
      FPArray<void *> *walk = data, *pointer;
      int size = _dimensions - 1, i;
      for ( i = 0; i < size; i++ )
        if ( ( pointer = (FPArray<void*>*) walk->element ( position [ i ] ) ) == 0 )
	  return NULL;
	else
	  walk = pointer;
      return (T*) walk->element ( position [ i ] );
    }
    inline bool exists ( INTTYPE *position, int length ) const {
      FPArray<void *> *walk = data, *pointer;
      int i;
      for ( i = 0; i < length; i++ )
        if ( ( pointer = (FPArray<void*>*) walk->element ( position [ i ] ) ) == 0 )
	  return false;
	else
	  walk = pointer;
      return true;
    }
    inline T* element ( INTTYPE *position ) const { return (*this) [ position ]; }
};

#endif
