/***************************************************************************
                          interpreter.cc  - interprets the facts in
			                    input files
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>
#include <string>
#include "interpreter.h"
#include "output.h"
#include "type.h"

using namespace std;

bool parsingbias = true;

bool interpret_predicate ( int size, AtomString **params ) {
  if ( size != 1 )
    return false;
  char *name = params[0]->name;
  size = params[0]->nrofparams;
  int *paramtypes = new int[size], typenr, nrofconst;
  params = params[0]->params;
  for ( int i = 0; i < size; i++ ) {
    if ( params[i]->nrofparams != 0 )
      return false;
    typetable.getType ( params[i]->name, typenr, nrofconst );
    if ( typenr == -1 )
      typenr = typetable.addType ( params[i]->name );
    paramtypes[i] = typenr;
  }
  predicatetable.addPredicate ( name, size, paramtypes );

  return true;
}

// Global data for chain modes.
int chainpos; // how deep in the chain is the current mode? Starts with 0
struct ChainMarker {
  int chainpos; // the position in the chain of the marker
  int paramnr; // which parameter at that position in the chain?
  ModeParamType type; // constant or output marker?
};
FArray<ChainMarker> *chainmarkers; // the chain markers.
FArray<ModeData*> *modehistory; // the history of parent modes in the chain tree.
ModeData *lastmode = NULL;
		  
void init_chain_modes () {
  if ( chainmarkers )
    delete chainmarkers;
  if ( modehistory )
    delete modehistory;
  chainmarkers = new FArray<ChainMarker>;
  modehistory = new FArray<ModeData*>;
}

bool interpret_startchain () {
  modehistory->insertAt ( chainpos, lastmode );
  chainpos++;
  return true;
}

bool interpret_endchain () {
  if ( chainpos == 0 ) {
    ERR << "There is no chain to end; ";
    return false;
  }
  chainpos--;
  return true;
}

bool interpret_mode ( int size, AtomString **params, char *functionname = NULL, bool iskey = false ) {
  if ( size != 1 )
    return false;
  bool chained = ( chainpos > 0 );
  ModeData *mdchainpred;
  
  if ( chained ) { // chain mode'
    mdchainpred = (*modehistory)[chainpos-1];
  }
  else {
    mdchainpred = NULL;
    init_chain_modes ();
  }
  int typenr, constantnr;
  char *name = params[0]->name;

  PredicateData *predicatedata = predicatetable.getPredicate ( name );
  if ( !predicatedata ) {
    ERR << "Unknown predicate " << name << "; ";
    return false;
  }
  if ( iskey )
    predicatedata->key = true;

  int predicate = predicatedata->nr,
      predsize = predicatedata->size;
  if ( params[0]->nrofparams != predsize ) {
    ERR << "Wrong number of parameters: ";
    return false;
  }
  if ( iskey ) {
    if ( predsize != 1 ) {
      ERR << "In the current version of Farmer a key predicate must have one argument; ";
      return false;
    }
    if ( params[0]->params[0]->name[0] != '-' ) {
      ERR << "The arguments of the key predicate can only be of mode `-'; ";
      return false;
    }
  }
  ModeParamData *modeparams = new ModeParamData[predsize];
  AtomString **walk = params[0]->params;

  for ( int i = 0; i < predsize; i++ )
    switch ( walk[i]->name[0] ) {
      case '+': case '=': {
        if ( walk[i]->name[1] != 0 ) {
	  int markernr = atoi ( walk[i]->name + 1 );
	  ChainMarker cm;
	  cm.chainpos = chainpos;
	  cm.paramnr = i;
	  cm.type = ChainPlus;
	  chainmarkers->insertAt ( markernr, cm );
	}
	if ( walk[i]->name[0] == '+' )
	  modeparams[i].type = Plus;
	else
	  modeparams[i].type = PlusEx;
        modeparams[i].number = predicatedata->params[i];
        break;
      }
      case '-': {
        if ( walk[i]->name[1] != 0 ) {
	  int markernr = atoi ( walk[i]->name + 1 );
	  ChainMarker cm;
	  cm.chainpos = chainpos;
	  cm.paramnr = i;
	  cm.type = ChainPlus;
	  chainmarkers->insertAt ( markernr, cm );
	}
        modeparams[i].type = MinEx;
        modeparams[i].number = predicatedata->params[i];
        break;
      }
      case '#': {
        if ( walk[i]->name[1] != 0 ) {
	  int markernr = atoi( walk[i]->name + 1 );
	  ChainMarker cm;
	  cm.chainpos = chainpos;
	  cm.paramnr = i;
	  cm.type = ChainConstant;
	  chainmarkers->insertAt ( markernr, cm );
	}
        modeparams[i].type = Constant;
	modeparams[i].number = predicatedata->params[i];
	// we replace this number in the final table with constant numbers.
	break;
      }
      default:
        if ( walk[i]->name[0] <= '0' || walk[i]->name[0] >= '9' ) {
	  ERR << "Unvalid " << i + 1 << "th mode parameter; ";
          return false;
        }
        int markernr = atoi( walk[i]->name );
	modeparams[i].back = (*chainmarkers) [ markernr ].chainpos - chainpos;
	modeparams[i].backparamnr = (*chainmarkers) [ markernr ].paramnr;
	modeparams[i].type = (*chainmarkers) [ markernr ].type;
	modeparams[i].number = predicatedata->params[i];
    }

  if ( functionname && strcmp ( functionname, "true" ) == 0 ) {
    lastmode = modetable.addMode ( predicate, predsize, modeparams, NULL, mdchainpred );
    lastmode->alwaystrue = true;
  }
  else {
    lastmode = modetable.addMode ( predicate, predsize, modeparams, functionname, mdchainpred );
    lastmode->alwaystrue = false;
  }
  if ( iskey )
    key = lastmode;

  if ( lastmode->alwaystrue && lastmode->nroutput ) {
    ERR << "You cannot define an atom which is always true when this atom has outputs; ";
    return false;
  }
  return true;
}

bool interpret_forbid ( int size, AtomString **params ) {
  if ( size != 2 ) {
    ERR << "The forbid bias specification atom requires 2 parameters; ";
    return false;
  }
  int typenr, constantnr;
  typetable.getType ( params[0]->name, typenr, constantnr );
  if ( typenr == -1 )
    typenr = typetable.addType ( params[0]->name );
  typetable.getConstant ( params[1]->name, typenr, constantnr );
  if ( constantnr == -1 )
    constantnr = typetable.addConstant ( params[1]->name, typenr );
  typetable.addForbid ( typenr, constantnr );
  return true;
}

bool interpret_predicatemax ( int size, AtomString **params ) {
  PredicateData *predicatedata = predicatetable.getPredicate ( params[0]->name );
  if ( size != 2 ) {
    ERR << "predicatemax requires a predicate and an integer as arguments; ";
    return false;
  }
  if ( !predicatedata ) {
    ERR << "Unknown predicate " << params[0]->name << "; ";
    return false;
  }
  predicatedata->number = atoi ( params[1]->name );
  return true;
}

bool interpret_fact ( char *name, int size, AtomString **params ) {
  PredicateData *predicatedata;
  if ( name[0] <= '9' )
    predicatedata = predicatetable.getPredicate ( atoi ( name ) );
  else
    predicatedata = predicatetable.getPredicate ( name );
  int typenr, constantnr;

  if ( !predicatedata ) {
    ERR << "Unknown predicate used in fact; ";
    return false;
  }
  if ( size != predicatedata->size ) {
    ERR << "Incorrect number of parameters; ";
    return false;
  }
  if ( predicatedata->key ) {
    ERR << "Can not use facts for the key; ";
    return false;
  }
  INTTYPE coor[size];
  for ( int i = 0; i < size; i++ ) {
    typenr = predicatedata->params[i];
    typetable.getConstant ( params[i]->name, typenr, constantnr );
    if ( constantnr == -1 )
      constantnr = typetable.addConstant ( params[i]->name, typenr );
    coor[i] = constantnr;
  }
  setFact ( predicatedata, coor );
  return true;
}

bool interpret_nooi ( int size, AtomString **params ) {
  int typenr, constantnr;
  if ( size != 1 ) {
    ERR << "nooi should have one argument; ";
    return false;
  }
  typetable.getType ( params[0]->name, typenr, constantnr );
  if ( typenr == -1 ) {
    ERR << "nooi should have a type as parameter; ";
    return false;
  }
  TypeData *td = typetable.typeData ( typenr );
  td->oi = false;
  return true;
}

bool interpret_pkey ( int size, AtomString **params ) {
  char *name = params[0]->name;
  PredicateData *predicatedata = predicatetable.getPredicate ( name );
  if ( !predicatedata ) {
    ERR << "Unknown predicate " << name << "; ";
    return false;
  }

  int predicate = predicatedata->nr,
      predsize = predicatedata->size;
  if ( params[0]->nrofparams != predsize ) {
    ERR << "Wrong number of parameters; ";
    return false;
  }
  AtomString **walk = params[0]->params;
  FArray<bool> *keyp = new FArray<bool>;

  for ( int i = 0; i < predsize; i++ )
    if ( walk[i]->name[0] != '_' )
      keyp->insertAt ( i, true );
  
  if ( !predicatedata->primarykeys )
    predicatedata->primarykeys = new FList<FArray<bool> *>;
  predicatedata->primarykeys->append ( keyp );
  
  return true;
}

bool interpret_minsup ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "minsup needs one integer parameter; ";
    return false;
  }
  minsup = atoi ( params[0]->name );
  return true;
}

bool interpret_showconstantnames ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "showconstantnames needs one parameter; ";
    return false;
  }
  if ( strcmp ( params[0]->name, "true" ) == 0 )
    showconstantnames = true;
  else
    if ( strcmp ( params[0]->name, "false" ) == 0 )
      showconstantnames = false;
    else {
      ERR << "showconstantnames parameter should be true or false; ";
      return false;
    }
  return true;
}

bool interpret_showstyle ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "showstyle needs one parameter; ";
    return false;
  }
  if ( strcmp ( params[0]->name, "queries" ) == 0 )
    showstyle = Queries;
  else
    if ( strcmp ( params[0]->name, "trie" ) == 0 )
      showstyle = Trie;
    else 
      if ( strcmp ( params[0]->name, "modetrie" ) == 0 )
        showstyle = ModeTrie;
      else {
        ERR << "showstyle must be queries, trie or modetrie; ";
        return false;
      }
  return true;
}

bool interpret ( AtomString *atom ) {
  if ( parsingbias ) {
    if ( strcmp ( atom->name, "predicate" ) == 0 )
      return interpret_predicate ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "predicatemax" ) == 0 )
      return interpret_predicatemax ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "forbid" ) == 0 )
      return interpret_forbid ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "mode" ) == 0 )
      return interpret_mode ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "b." ) == 0 )
      return interpret_startchain ();
    if ( strcmp ( atom->name, "e." ) == 0 )
      return interpret_endchain ();
    if ( strcmp ( atom->name, "externmode" ) == 0 )
      return interpret_mode ( atom->nrofparams - 1, atom->params + 1, atom->params[0]->name );
    if ( strcmp ( atom->name, "key" ) == 0 )
      return interpret_mode ( atom->nrofparams, atom->params, NULL, true );
    if ( strcmp ( atom->name, "pkey" ) == 0 )
      return interpret_pkey ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "nooi" ) == 0 )
      return interpret_nooi ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "minsup" ) == 0 )
      return interpret_minsup ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "showstyle" ) == 0 )
      return interpret_showstyle ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "showconstantnames" ) == 0 )
      return interpret_showconstantnames ( atom->nrofparams, atom->params );
    parsingbias = false;
      // no bias definition, going to background knowledge.
      // first put bias in faster access datastructures
    ERR << "[ Initializing fact tables ]" << endl;
    ERR_F
  }

  return interpret_fact ( atom->name, atom->nrofparams, atom->params );
}
