/***************************************************************************
                          modetable.h  -  the dictionary of modes
                             -------------------
    begin                : Wed Mar 22 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef MODETABLE_H
#define MODETABLE_H
#include "typetable.h"
#include "facttable.h"
#include "tree.h"
#include "type.h"
#include "farray.h"
#include "modefacttable.h"

extern TypeTable typetable;

class PredicateTable;
class PredicateData;

enum ModeParamType { Plus, ChainPlus, PlusEx, Min, MinEx, Constant, ChainConstant };
typedef unsigned int ModePatternType;
inline void init ( ModePatternType &p ) {
  p = 0;
}
inline void setOn ( ModePatternType &p, int n ) {
  p |= ( 1 << n );
}
inline bool is ( ModePatternType p1, ModePatternType p2 ) {
  return p1 == p2;
}

#define IS_VAR_PARAM(param) ( (param)->type <= MinEx )
#define IS_CONST_PARAM(param) ( (param)->type == Constant )
#define IS_OUTPUT_PARAM(param) ( (param)->type == MinEx )
#define IS_INPUT_PARAM(param) ( (param)->type == Plus || \
                                (param)->type == PlusEx || \
                                (param)->type == Constant ||  \
				(param)->type == ChainPlus || \
				(param)->type == ChainConstant )

struct ModeParamData {
  ModeParamType type;
  int number;
  
  // only used by chained plus:
  short int back;  // how many atoms back is the variable introduced?
  short int backparamnr; // at which position?
};

struct ModeData {
  int nr;                 // position in the modetable
  int deepnr;             // position in the deeptable with given modes.
  int firstnr;            // position in the modetable of the first mode generated by the same original deeparray mode.
  int predicate;          // predicate number
  bool haschainsucc;      // is there a next mode in the chain in the deep table?
  bool haschainpred;      // is there a previous mode in the chain in the deep table?
  bool haschainconst;     // does this mode contain a chained constant?
  FArray<ModeData*> chainedmodes; // the modes which are chained to this mode,
                            // this are indexes in the deep table.
  int chainprednr;       // if this mode is in a chain, we store
                          // the predecessor mode (in deeptable) to identify the block.
			  // Note that this only makes sence if haschainpred==true
  int chaindeepnr;        // == deepnr if there is chained mode, otherwise -1			  
  int minchainlength;     // the lenght of the smallest chain starting from this node
  ModeParamData *params;  // parameters
  int size;               // nr of parameters
  int *dependentmodes;    // pointers to the first mode generated by an original dependent mode.
  ModePatternType pattern; // a binary pattern with a 1 at input positions.
  FArray<int> equivalentmodes; // pointers to first modes which may cause equivalent atoms.
                               // be careful about constants. For the given result, add
			       // (nr-firstnr) of the mode for which equivalence is tested.
			       // This is due to the fixed order of summing constants.
  int depmodsize;         // nr of such pointers
  int *inputvarpos,       // input variable positions, for fast walking
      *outputvarpos;
  int nrinput, nroutput;  // nr of input variables
  int nrconst;            // nr of constants. Used as a quick heuristic to check equivalency
  int nrplusex;           // nr of plusex variables.
  bool outputs;           // has outputs?
  bool equalingmode;      // if no, is there a mode for this predicate
                          // which does at at least one position where
                          // this mode has an input?
  bool alwaystrue;        // should this node always be evaluated to true?
                          // this is only allowed for nodes with only inputs,
			  // occuring in chains and should be used with very much
			  // care.
  int nrofmodes;          // the number of modes after this mode that have been generated by the same original mode
  bool quickpreeval;      // using the facttable and modefacttable, determine if
                          // it is possible to quickly determine whether
			  // a fact occurs for some inputs. Otherwise recursion
			  // can be avoided.
  int quickpreevalsize;   // the number of parameters to be specified to allow
                          // quick preevaluation.
  union {
    BaseFactTable *facttable; // if no outputs, binary matrix
    BaseModeFactTable *modefacttable;
                          // otherwise, a matrix containing true output values
  } table;
  bool reuses;            // is another table reused?
  char *functionname;
  FList<FArray<int> *> *primarykeys;
  
  ActiveList activefrequent; // determines which modes in the
                             // complete table have been found
			     // frequent 1-item sets.
			     // Allows for the quick discovery
			     // of useful modes.
};


// only used internally
struct ModeConstData {
  int nr;
  ModeData *data;
  ModeParamData *params; // parameters with filled in constants
};

// datastructures explanation:
// There is a `deep table' which contains  ModeData structs. This structs store the
// modes as defined by the user. Some of these modes contain `#'. This `#' is replaced by
// a constant in the preprocessing phase. A new table is constructed (array), which contains
// ModeConstData structs. Each of these structs contains a pointer to the ModeData that
// generated this ModeConstData and a set of parameters (in which the `#' is replaced by a
// constant).

void setFact ( ModeData *md, INTTYPE *coors );

class ModeTable {
  public:
    ModeTable ();
    ~ModeTable ();
    int cdepsize;
    ModeData* addMode ( int predicate, int parsize, ModeParamData *params, char *functionname, ModeData *mdchainpred );
    void print ();
    void fillDependencies ();
    int searchEquivalentMode ( Node *node, ModePatternType m );
    inline ModeData *modeData ( int nr ) {
      ModeConstData &mcd = array[nr];
      ModeData *md = mcd.data;
      md->params = mcd.params;
      md->nr = mcd.nr;
      return md;
    }
    inline int chainprednr ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->chainprednr;
    }
    inline int haschainsucc ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->haschainsucc;
    }
    inline int haschainpred ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->haschainpred;
    }
    inline int doesonechainfinish ( int remainingdepth, int modenr ) {
      ModeData *md = array[modenr].data;
      if ( md->haschainsucc )
        return remainingdepth >= md->minchainlength;
      else
        return true;
    }
    inline bool quickpreeval ( int modenr ) const {
      return array[modenr].data->quickpreeval;
    }
    inline int quickpreevalsize ( int modenr ) const {
      return array[modenr].data->quickpreevalsize;
    }
    inline bool alwaystrue ( int modenr ) const {
      return array[modenr].data->alwaystrue;
    }
    inline bool computeQuickPreEval ( int modenr, int keyval, INTTYPE *params ) const {
      ModeData *md = array[modenr].data;
      int size = md->quickpreevalsize;
      if ( md->table.modefacttable ) {
        bool val;
	int temp = params[0];
	params[0] = keyval;
	if ( md->nroutput  )
          val = ((ModeFactTable*) md->table.modefacttable)->existsFact ( params, size );
	else // table is a union.
          val = ((FactTable*) md->table.facttable)->existsFact ( params, size );
	params[0] = temp;
	return val;
      }
      else
        return false;
    }
    inline int chaindeepnr ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->chaindeepnr;
    }
    inline ModeData *nextFrequentActive ( int &modenr ) {
      ModeData *md2 = array[modenr].data;
      modenr = md2->activefrequent.nextActive ( modenr - md2->firstnr ) + md2->firstnr;
      return modeData ( modenr );
    }
    inline int *inputVarPos ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->inputvarpos;
    }
    inline int *outputVarPos ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->outputvarpos;
    } 
    inline ModeParamData *params ( int modenr ) const {
      return array[modenr].params;
    }
    inline int outputType ( int modenr, int outputpos ) const {
      return array[modenr].params[outputpos].number;
    }
    inline ModeParamType modeParamType ( int modenr, int param ) const {
      return array[modenr].params[param].type;
    }
    inline int nrParams ( int modenr ) const {
      return array[modenr].data->size;
    }
    inline int nrOutput ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->nroutput;
    }
    inline ModeFactTable *modefacttable ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return (ModeFactTable*) md->table.modefacttable;
    }
    inline bool alwaysTrue ( int modenr ) const {
      return array[modenr].data->alwaystrue;
    }
    inline BaseFactTable *facttable ( int modenr ) const {
      return  (BaseFactTable*) array[modenr].data->table.facttable;
    }
    inline double outputListLength ( int modenr ) const {
      ModeData *md = array[modenr].data;
      if ( md->nroutput && ((ModeFactTable*) md->table.modefacttable) ) 
        return ((ModeFactTable*) md->table.modefacttable)->meanListLength ();
      else
        return 0.0;
    }
    inline int nrConst ( int modenr ) const {
      return array[modenr].data->nrconst;
    }
    inline int outputs ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->outputs;
    }
    inline int nrInput ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->nrinput;
    }
    inline int deepmode ( int modenr ) const {
      ModeData *md = array[modenr].data;
      return md->deepnr;
    }
    FArray<ModeData*> &chainedmodes ( int deepmodenr ) const {
      return deeparray[deepmodenr]->chainedmodes;
    }
    inline int size () const { return _size; }
    ModeData *chaingoback ( ModeData *md, int times ) const {
      for ( int i = 0; i < times; i++ )
        md = deeparray[md->chainprednr];
      return md;
    }
    inline int deepsize () const { return deeparray.size (); }
    void inactivateAllEquivalent ( Node *node );
    bool variablevsconstantcheck;
  private:
    void checkNOOI ();
    bool checkNOOIMode ( int modenr );
    void determineQuickPreEval ( ModeData *md );
    void determinePKeys ( ModeData *md, PredicateData *pd );
    void createArray ();
    void createTableForMode ( ModeData *md );
    void prepend ( ModeData *md );
    void fillAllModes ( ModeParamData *params, ModeData *md, int &fillsize, int position, ModeParamData *mpd );
    void fillMinChainLength ( ModeData *fillmd, ModeData *walkmd, int length );
    FArray<ModeData*> deeparray; // contains all modes as defined.
    ModeConstData *array; // contains all modes, including expanded constant modes.
    int _size; // size of the latter array
    bool filledarray;
};

extern ModeTable modetable;

#endif
