/***************************************************************************
                          tree.h - declares all classes involved
			           in evaluating and generating 
				   queries in a trie
                             -------------------
    begin                : Sat Mar 25 2000
    copyright            : (C) 2000-2003 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef TREE_H
#define TREE_H
#include "type.h"
#include "farray.h"
#include "flist.h"

class ChildList;
class TidList;

extern int keepequivalent;

struct Node {
  INTTYPE mode;
  INTTYPE *params;
  unsigned int support;
    // support = '-1' if the node is equivalent to another
    // node but could not be removed to guarantee completeness of search.
  unsigned int backtracksize;
  ChildList *freqchildren;
  ChildList *infreqchildren;
  TidList *tidlist;
  char maxdepth; // the longest query in the subtree below this query.
  //char introdepth; // we will store the depth at which this node was introduced to allow
                   // for equivalency checking and signalling.
};

#define MODE(node) node->mode

// order used when storing nodes
int nodeCmp ( Node *node1, Node *node2 );

class HashNodes {
  public:
    FArray<Node> nodes; // we allocate exactly the required amount of space.
    HashNodes ( int size ): nodes ( size ) {  }
};

class ChildList;

class TreeNodePointer {
  private:
  public:
    Node *node; // this node...
    ChildList *childlist; // ...occurs in this childlist...
    int hashnr; // ...with this mode...
    int pos; // ... at this position 
};

// order used while generating (expanding) nodes
int nodeCmp2 ( Node *node1, Node *node2 );
int nodeCmp2b ( TreeNodePointer *tnp1, TreeNodePointer *tnp2 );
int nodeCmp3 ( const void *tnp1, const void *tnp2 );

class ChildList {
  static TreeNodePointer searchresult;
  public:
    ChildList ();
    ~ChildList ();
    // only fills hashnr and pos.
    TreeNodePointer *search ( Node *node );
    TreeNodePointer *searchbeforeequal ( Node *node, TreeNodePointer *before );   
    FArray<Node> &allocate ( int modenr, int size );
    FArray<Node> &nodes ( int modenr ) { return (*hash) [modenr-modefirstchild]->nodes; }
    FArray<Node> *nodess ( int deepmodenr );
    FArray<HashNodes*> *hash;
    int modefirstchild;
};

struct ListModeFactCoorsElement;

class TidData {
  public:
    TidData ( int tid ) : tid ( tid ), assignments ( NULL ) { }
    TidData (): tid ( -1 ), assignments ( NULL ) { }
    int tid;
    FArray<ListModeFactCoorsElement*> *assignments; 
    void print ();
    inline int size () const { if ( assignments ) return assignments->size (); else return 0; }
    ~TidData ();
    // for the moment we do not allow external functions for predicates with
    // an output!!!
};

class TidList {
  public:
    void print ();
    FList<TidData> tids;
    int refcount;
    unsigned int backtracksize;
    void kill () { refcount--; if ( refcount == 0 ) delete this; }
    TidList *copy () { refcount++; return this; }
    TidList (): backtracksize ( 0 ), refcount ( 1 ) { }
};

class FullTidData {
  public:
    TidData changes;
    FArray<short int>** assignments;
    FullTidData ( int tid );
    inline int tid () const { return changes.tid; }
    ~FullTidData ();
    bool checkConsistent () const; // for debugging purposes
    void apply ( TidData &tidd, int backinhistory, int backintidd );
    TidData *undoApplyData ( TidData &tidd, int back, int backintidd );
    void applyToAssignments ( Node *node, ListModeFactCoorsElement *el );
};

class FullTidArray {
  public:
    FArray<FullTidData*> tids;
    FullTidArray ();
    ~FullTidArray ();
};

extern FullTidArray *fulltidarray;

class NodeHistory {
  private:
    int top;
    FArray<TreeNodePointer> elements;
    // query information 
    int *variables;
    bool **constappears;
    FArray<FArray<bool>*> constappearshistory;
    FArray<int> outputatoms;
    int nroutputatoms;
    FArray<bool> equivalentpossible;
  public:
    FArray<FArray<TreeNodePointer>*> nodeorders;
    FArray<int>** variableoutputatom;
    FArray<TidList*> assignundos;
    NodeHistory ();
    TreeNodePointer &last () { return elements[top]; }
    inline bool equivalentPossible ( int i ) const { return equivalentpossible[i]; }
    void init ();
    inline TreeNodePointer &element ( int i ) const { return elements[i]; }
    Node *outputNode ( int i ) const { // 0 == atom AFTER the root
      return elements[outputatoms[i+1]].node;
    }
    inline TreeNodePointer back ( int i ) const { return elements[top + i]; }
    void push ( TreeNodePointer &tnp );
    TreeNodePointer &pop ();
    inline int size () const { return top + 1; }
    void print ();
    void print2 ();
    // Information about the query constructed this far:
    inline int *queryVariableCounts () const {
      return variables;
    }
    inline bool queryConstAppears ( int typenr, int varnr ) const {
      return constappears[typenr][varnr];
    }
    inline int queryNrOutputAtoms () const {
      return nroutputatoms;
    }
    inline FArray<int>& queryOutputAtoms () {
      return outputatoms;
    }
};

extern NodeHistory nodeHistory;

class ActiveList {
  private:
    FArray<int> prev;
    FArray<int> next;
    int firstactive, activesize;
  public:
    ActiveList ();
    void setActive ( int size ); // sets size items to active.
    void setInActive ( int pos ); // sets the pos-th item to inactive
    inline int nextActive ( int pos ) const {
      if ( next[pos] == -1 )
        return -1;
      return next[pos] + pos + 1; // goes to the next active from pos
    }
    inline int firstActive () const {
      return firstactive; // goes the the first active.
    }
    int activeSize () const {
      return activesize;
    }
};

struct ModeData;

class QueryExpansions {
  public:
    ChildList candidates;
    ActiveList activeactivelist; // to store which buckets are active.
    ActiveList* activelist; // pointer to an array of active lists. For each bucket an array.
    FArray < FListWalk < TidData > >* brothertidwalk; // for each node in the candidates,
      // if this candidate is a copy of a brother, we walk through the brother's tidlist too.
      // this is the set of walkers for brothers. An array element == NULL if this is not a copy
      // of a brother.
      // For each bucket an array.
    int currentbucket, currentpos;
    int nextbucket, nextpos;

    bool alwaystrue;
    int initactivesize;

    QueryExpansions ();
    ~QueryExpansions ();
    void init (); // call before adding new nodes.
    void addNode ( ModeData *md, Node *node );
    void setTid ( int tid );
    void firstActive ();
    Node *currentActive ();
    void setCurrentInActive ();
    void nextActive ();
    void initialize ();
};

// evaluates a query which is stored into the node history.
class QueryEvaluator {
  private:
    ListModeFactCoorsElement *evaluateOutputs ( Node *node );
    inline bool evaluateNoOutputs ( Node *node );
    int constAppears ( int typenr, int valnr );
    int validAssignment ( Node *node, INTTYPE *value );
    inline int validConstants ( Node *node );
    ListModeFactCoorsElement *findFirstValidAssignment ( Node *node, ListModeFactCoorsElement *startcoors );
    void determineBacktrackPos ( Node *node, ListModeFactCoorsElement *startcoors );
    void addTid ( Node *node, int btstartpos );
    bool evaluateLastOutputs ( Node *node );
    bool evaluateLastNoOutputs ( Node *node );
    void backtrackPosOnInputs ( Node *node );
    void backtrackPosOnInputs2 ( Node *node, int pos );
    bool evaluateExpansions ();
    void addToNrVariables ( Node *node );
    void removeFromNrVariables ( Node *node );
    void undoAssignments ();
    ListModeFactCoorsElement* element ( int i );
    void set ( int i, ListModeFactCoorsElement *el );
    bool evaluate_nooutputs ( int outputpos );
    FullTidData *ftd;
    QueryExpansions *queryexpansions;
    TidData applytid;
    INTTYPE *coors;
    int *nrvariables;
    int firstchangepos;
    FArray<bool> backtrackstop;
  public:
    QueryEvaluator ();
    ~QueryEvaluator ();
    void printVariables ();
    void evaluate ( FullTidData *ftd, QueryExpansions *queryexpansions );
    void initialize ();
};

void print ();
void fillTidDependent ();
void fillTidBrother ( Node *orignode );
void applyAllAssignments ();
void undoAllAssignments ();
void printNode ( Node *node );
bool frequentTidUnion ( TidList *list1, TidList *list2 );
bool searchforequalnode ( Node *newnode );
bool searchforequivalentquery ();

#include "modefacttable.h"

#endif
