/***************************************************************************
                          treecount.cpp  - evaluates a set of
			                   candidates against the
					   database
                             -------------------
    begin                : 26 Mar 2003
    copyright            : (C) 2003 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "tree.h"
#include "globaldata.h"
#include "output.h"
#include <string.h>
#include <values.h>

ActiveList::ActiveList () {
}

void ActiveList::setActive ( int size ) {
   // sets size items to active.
   prev.resize ( 0 );
   next.resize ( 0 );
   if ( size ) {
     prev.insertAt ( size - 1, 0 ); // use the zeroing mechanisms of FArrays
     next.insertAt ( size - 1, -1 );
     firstactive = 0;
     activesize = size;
   }
   else {
     firstactive = -1;
     activesize = 0;
   }
}

void ActiveList::setInActive ( int pos ) {
   if ( next[pos] == -2 )
     return;
   // sets the pos-th item to inactive
   if ( pos == firstactive )
     if ( next[pos] == -1 )
       firstactive = -1;
     else
       firstactive = pos + next[pos] + 1;
   else {
     int previ = pos - prev[pos] - 1;
     if ( next[pos] == -1 )
       next[previ] = -1;
     else
       next[previ] = pos - previ + next[pos];
   }
   if ( next[pos] != -1 ) {
     int nexti = pos + next[pos] + 1;
     prev[nexti] = nexti - pos + prev[pos];
   }
   next[pos] = -2; // mark as inactive.
   activesize--;
}

FullTidData::FullTidData ( int tid ) {
  static INTTYPE value;
  static ListModeFactCoorsElement el = { &value, NULL };
  value = changes.tid = tid;
  
  assignments = new FArray<short int>* [ typetable.size () ];
  for ( int i = 0; i < typetable.size (); i++ )
    assignments[i] = new FArray<short int>;
  
  applyToAssignments ( nodeHistory.element ( 0 ).node, &el );
}

FullTidData::~FullTidData () {
  for ( int i = 0; i < typetable.size (); i++ )
    delete assignments[i];
  delete[] assignments;
}

bool FullTidData::checkConsistent () const {
  for ( int i = 0; i < changes.size (); i++ ) {
    ListModeFactCoorsElement *el = (*changes.assignments)[i];
    Node *node = nodeHistory.element ( nodeHistory.queryOutputAtoms () [ i + 1 ] ).node;
    ModeData *md = modetable.modeData ( node->mode );
    for ( int k = 0; k < md->nroutput; k++ ) {
      int val = (*assignments[md->params[md->outputvarpos[k]].number])[node->params[md->outputvarpos[k]]];
      if ( val != el->value[k] ) {
	ERR << "Consistency error. Tid " << changes.tid << " Node: ";
	printNode ( node );
	ERR << " Type " << md->params[md->outputvarpos[k]].number <<
	       " Variable " << node->params[md->outputvarpos[k]]
	    << " Value " << val << " =/= " << el->value[k] << endl;
        ERR_F
	return false;
      }
    }
  }
  return true;
}

void TidData::print () {
  int typenr = key->params[0].number;
  cout << typetable.getConstantName ( typenr, tid );
  if ( assignments ) {
    for ( int j = 0; j < assignments->size (); j++ ) {
      ListModeFactCoorsElement *el = (*assignments) [ j ];
      //cout << ":" << *el->value;
    }
  }
}

void TidList::print () {
  FListWalk<TidData> walk ( &tids );
    
  if ( tids.size () == 0 ) {
    cout << "[]";
    return;
  }
  
  cout << " [";
  
  for ( int i = 0; i < tids.size () - 1; i++ ) {
    walk.current ()->print ();
    cout << ",";
    walk.next ();
  }
  walk.current ()->print ();
  cout << "]";
}

void FullTidData::applyToAssignments ( Node *node, ListModeFactCoorsElement *el ) {
  INTTYPE *value = el->value;
  int *outputvarpos = modetable.outputVarPos ( node->mode );
  int size = modetable.nrOutput ( node->mode );
  for ( int j = 0; j < size; j++ ) {
    int outputpos = outputvarpos[j];
    assignments[modetable.outputType ( node->mode, outputpos )]->insertAt
      ( node->params[outputpos], value[j] );
  }
}

// apply the given TidData struct starting in the current
// tid from the back-th position from the back,
// and in the TidData from the bachintidd.
void FullTidData::apply ( TidData &tidd, int back, int backintidd ) {
  int index = changes.size () - back - 1;
  if ( tidd.size () > 0 && !changes.assignments )
    changes.assignments = new FArray<ListModeFactCoorsElement*>;
  for ( int i = backintidd; i < tidd.size (); i++, index-- ) {
    ListModeFactCoorsElement *el = (*tidd.assignments)[i];
    applyToAssignments ( nodeHistory.outputNode ( index ), el );
    changes.assignments->insertAt ( index, el );
  }
  for ( int k = 0; k < back; k++ )
    changes.assignments->releaselast ();
}

// determine an undo struct for the apply by the (tidd,back,backintidd) pair.
// In this undo struct, the first back places are left empty.
// This space should be filled by the query evaluation class.
// Note: the the last atom of a query needs never be undone.
TidData *FullTidData::undoApplyData ( TidData &tidd, int back, int backintidd ) {
  static TidData tid;
  tid.tid = tidd.tid;
  tid.assignments = new FArray<ListModeFactCoorsElement*>;
  
  int size = tidd.size () - backintidd + back;
  
  int walk = changes.size () - 1;
  for ( int k = 0; k < size; k++, walk-- )
    tid.assignments->insertAt ( k, (*changes.assignments)[walk] );

  for ( int k = 0; k < back; k++ )
    tid.assignments->insertAt ( k, (*changes.assignments)[changes.size () - k - 1] ); // +1 start of array at second atom
  
  return &tid;
}

FullTidArray::FullTidArray () {
  ModeData *md = key;
  TypeData *td = typetable.typeData ( key->params[0].number );

  for ( int i = 0; i < td->constnumbers; i++ )
    tids.insertAt ( i, new FullTidData ( i ) );
}

FullTidArray::~FullTidArray () {
  for ( int i = 0; i < tids.size (); i++ )
    delete tids[i];
}

QueryEvaluator::QueryEvaluator () {
}

void QueryEvaluator::initialize () {
  applytid.assignments = new FArray<ListModeFactCoorsElement*>;
  coors = new INTTYPE[predicatetable.maxNParams ()];
  nrvariables = new int[typetable.size ()];
}

QueryEvaluator::~QueryEvaluator () {
  delete[] coors;
  delete[] nrvariables;
}


ListModeFactCoorsElement *QueryEvaluator::evaluateOutputs ( Node *node ) {
  ModeData *md = modetable.modeData ( node->mode );
  int t = 0;
  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case Plus: case ChainPlus: case PlusEx:
        coors[t] = (*ftd->assignments[md->params[i].number])[node->params[i]];
	t++;
	break;
      case Constant: case ChainConstant:
        coors[t] = node->params[i];
	t++;
    }
  }
  return ((ModeFactTable*) md->table.modefacttable)->getListFacts ( coors );
}

inline bool QueryEvaluator::evaluateNoOutputs ( Node *node ) {
  if ( modetable.alwaysTrue ( node->mode ) )
    return true;
  ModeData *md = modetable.modeData ( node->mode );
  INTTYPE *coorsp = coors;
  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case Plus: case ChainPlus: case PlusEx:
        *coorsp = (*ftd->assignments[md->params[i].number])[node->params[i]];
	coorsp++;
	break;
      case Constant: case ChainConstant:
        *coorsp = node->params[i];
	coorsp++;
    }
  }
  return md->table.facttable->getFact ( coors  );
}

inline void QueryEvaluator::backtrackPosOnInputs ( Node *node ) {
  ModeData *md = modetable.modeData ( node->mode );
  int *inputvarpos = md->inputvarpos;
  int pos;

  //int maxpos = -1;
  int val;

  for ( int i = 0; i < md->nrinput; i++ ) {
    pos = inputvarpos[i];

    val = (*nodeHistory.variableoutputatom[md->params[pos].number])[node->params[pos]];
    //maxpos = max ( val, maxpos );
    backtrackstop[val] = true;
  }

  //backtrackstop[maxpos] = true;
}

inline void QueryEvaluator::backtrackPosOnInputs2 ( Node *node, int pos ) {
  ModeData *md = modetable.modeData ( node->mode );
  int *inputvarpos = md->inputvarpos;
  int pos2;

  int maxpos = -1;
  int val;

  for ( int i = 0; i < md->nrinput; i++ ) {
    pos2 = inputvarpos[i];

    val = (*nodeHistory.variableoutputatom[md->params[pos2].number])[node->params[pos2]];
    maxpos = max ( val, maxpos );
    backtrackstop[val] = true;
  }

  for ( maxpos++; maxpos < pos; maxpos++ )
    backtrackstop[maxpos] = false;
  //backtrackstop[maxpos] = true;
}

int QueryEvaluator::constAppears ( int typenr, int valnr ) {
  // is this constant assigned to a variable?
  FArray<short int> &assignments = *(ftd->assignments[typenr]);

  for ( int i = 0; i < nrvariables[typenr]; i++ )
    if ( assignments[i] == valnr )
      return ( *nodeHistory.variableoutputatom[typenr] ) [ i ];

  return INT_MAX;
}

void QueryEvaluator::printVariables () {
  for ( int i = 0; i < typetable.size (); i++ ) {
    FArray<short int> &assignments = *(ftd->assignments[i]);
    cout << "Type " << i << ":";
    for ( int j = 0; j < nrvariables[i]; j++ )
      cout << assignments[j] << ";";
  }
}

int QueryEvaluator::validAssignment ( Node *node, INTTYPE *value ) {
  // is the given assignment to output variables in node valid,
  // which means those values do not occur in all atoms except (perhaps) the last one,
  // and is not assigned to any variables except (perhaps) the last one.
  int nroutput = modetable.nrOutput ( node->mode );
  int *outputvarpos = modetable.outputVarPos ( node->mode );
  int mini = INT_MAX;
  int minval;

  for ( int i = 0; i < nroutput; i++ ) {
    int typenr = modetable.outputType ( node->mode, outputvarpos[i] );
    int valnr = value[i];

    if ( typetable.typeData ( typenr )->oi ) {

      if ( nodeHistory.queryConstAppears ( typenr, valnr ) ) // check the atom itself! (May not be in history)
       // (occurs as constant in query) BUG!!!!
       // NOT ALLOWED NOW: CONSTANT and OUTPUT of the same type in the same atom!
        return 0;
      minval = constAppears ( typenr, valnr );
        // check that outputs do not have the same value! BUG!!!
       // (is assigned to variable)
       // NOT ALLOWED NOW: TWO or MORE OUTPUTS with the same type in the same atom!
      mini = min ( minval, mini );
    }
  }
  return mini;
}

inline int QueryEvaluator::validConstants ( Node *node ) {
  // given the current assignment of variables, are the constants valid?
  int mini = INT_MAX;
  if ( modetable.variablevsconstantcheck && modetable.nrConst ( node->mode ) != 0 ) {
    ModeData *md = modetable.modeData ( node->mode );
    PredicateData *pd = predicatetable.predicateData ( md->predicate );
    int minval;

    for ( int i = 0; i < pd->size; i++ ) {
      if ( md->params[i].type == Constant || md->params[i].type == ChainConstant ) {
        // search in assignments for this type.
        if ( typetable.typeData ( md->params[i].number )->oi ) {
          minval = constAppears ( pd->params[i], node->params[i] );
          mini = min ( minval, mini );
	}
      }
    }
  }
  return mini;
}

ListModeFactCoorsElement *QueryEvaluator::findFirstValidAssignment ( Node *node, ListModeFactCoorsElement *startcoors ) {
  ListModeFactCoorsElement *walk = startcoors;
  int minval;

  while ( walk ) {
    minval = validAssignment ( node, walk->value );
    if ( minval == INT_MAX )
      break;
    backtrackstop[minval] = true;
    walk = walk->next;
  }

  return walk;
}

void QueryEvaluator::determineBacktrackPos ( Node *node, ListModeFactCoorsElement *startcoors ) {
  ListModeFactCoorsElement *walk = evaluateOutputs ( node );
  int minval;

  while ( walk != startcoors ) {
    minval = validAssignment ( node, walk->value );
    if ( minval != INT_MAX )
       backtrackstop[minval] = true;
    walk = walk->next;
  }
}

void QueryEvaluator::addTid ( Node *node, int btstartpos ) {
  TidData newtidd;
  newtidd.tid = ftd->changes.tid;
  int as = applytid.assignments->size () - btstartpos - firstchangepos;
  if ( as > 0 ) {
    newtidd.assignments = new FArray<ListModeFactCoorsElement*> ( as );
    for ( int k = 0, l = applytid.assignments->size () - btstartpos - 1; k < as; k++, l-- )
      newtidd.assignments->insertAt ( k, (*applytid.assignments) [ l ] );
  }
  else
    newtidd.assignments = NULL;
  if ( !node->tidlist )
    node->tidlist = new TidList;
  node->tidlist->tids.append ( newtidd );
  node->tidlist->backtracksize += as;
  newtidd.assignments = NULL; // prevent deleting
}

bool QueryEvaluator::evaluateLastOutputs ( Node *node ) {
  int backtrackpos = validConstants ( node );
  if ( backtrackpos == INT_MAX ) {
    ListModeFactCoorsElement *el = evaluateOutputs ( node );
    if ( !el )
      return false;
    el = findFirstValidAssignment ( node, el );
    if ( el ) {
      // we have found an assignment.
      if ( ftd->changes.assignments )
        applytid.assignments->insertAt ( ftd->changes.assignments->size (), el );
      else
        applytid.assignments->insertAt ( 0, el );
      addTid ( node, 0 );
      return true;
    }
  }
  else
    backtrackstop[backtrackpos] = true;
  return false;
}

bool QueryEvaluator::evaluateLastNoOutputs ( Node *node ) {
  int backtrackpos = validConstants ( node) ;
  if ( backtrackpos == INT_MAX ) {
    if ( evaluateNoOutputs ( node ) ) {
      // we have found an assignment.
      addTid ( node, 1 );
      return true;
    }
  }
  else
    backtrackstop[backtrackpos] = true;

  return false;
}

bool QueryEvaluator::evaluateExpansions () {
  queryexpansions->firstActive ();
  Node *node;
  bool ret = false;

  while ( node = queryexpansions->currentActive () ) {
    if ( modetable.nrOutput ( node->mode ) ) {
      if ( evaluateLastOutputs ( node ) )
        queryexpansions->setCurrentInActive ();
      else {
        backtrackPosOnInputs ( node );
	ret = true;
      }
    }
    else {
      if ( evaluateLastNoOutputs ( node ) )
        queryexpansions->setCurrentInActive ();
      else {
        backtrackPosOnInputs ( node );
	ret = true;
      }
    }
    queryexpansions->nextActive ();
  }

  return ret;
}

void QueryEvaluator::addToNrVariables ( Node *node ) {
  ModeData *md = modetable.modeData ( node->mode );
  int position;
  for ( int i = 0; i < md->nroutput; i++ ) {
    position = md->outputvarpos[i];
    nrvariables[md->params[position].number]++;
  }
}

void QueryEvaluator::removeFromNrVariables ( Node *node ) {
  ModeData *md = modetable.modeData ( node->mode );
  int position;
  for ( int i = 0; i < md->nroutput; i++ ) {
    position = md->outputvarpos[i];
    nrvariables[md->params[position].number]--;
  }
}

void QueryEvaluator::undoAssignments () {
  for ( int r = firstchangepos; r < applytid.size () - 1; r++ ) // the last doesn't need to be undone.
    ftd->applyToAssignments ( nodeHistory.outputNode ( r ),
                              (*ftd->changes.assignments)[r] );
}

ListModeFactCoorsElement* QueryEvaluator::element ( int i ) {
  if ( i >= firstchangepos )
    return (*applytid.assignments) [ i ];
  else
    return (*ftd->changes.assignments) [ i ];
}

void QueryEvaluator::set ( int i, ListModeFactCoorsElement *el ) {
  (*applytid.assignments) [ i ] = el;
  if ( i < firstchangepos )
    firstchangepos = i;
}

bool QueryEvaluator::evaluate_nooutputs ( int outputpos ) {
  int startposition = nodeHistory.queryOutputAtoms () [ outputpos + 1 ] + 1; // .. + 1 ]: key in history
  int endposition;
  if ( outputpos + 2 == nodeHistory.queryNrOutputAtoms () ) // +1 additional as key in history
    endposition = nodeHistory.size ();
  else
    endposition = nodeHistory.queryOutputAtoms () [ outputpos + 2 ]; // + 1 key in history
  for ( int k = startposition; k < endposition; k++ ) {
    if ( !evaluateNoOutputs ( nodeHistory.element ( k ).node ) ) {
      backtrackPosOnInputs2 ( nodeHistory.element ( k ).node, outputpos );
      return false;
    }
  }
  return true;
}

void QueryEvaluator::evaluate ( FullTidData *ftd, QueryExpansions *queryexpansions ) {
  this->ftd = ftd;
  this->queryexpansions = queryexpansions;
  int outputpos = ftd->changes.assignments ? ftd->changes.assignments->size () : 0;
  firstchangepos = outputpos;
  int startpos = outputpos;
  int minpos = outputpos;
  ListModeFactCoorsElement *el, *el2;
  applytid.assignments->resize ( outputpos + 1 );
  backtrackstop.resize ( 0 );
  backtrackstop.insertAt ( outputpos, false );

  memcpy ( nrvariables, nodeHistory.queryVariableCounts (), sizeof(int) * typetable.size () );
  Node *node;

  if ( evaluateExpansions () ) {
    if ( outputpos <= 0 ) {
        undoAssignments ();
        return;
    }

    node = nodeHistory.outputNode ( outputpos - 1 );
    removeFromNrVariables  ( node );
    el = element ( outputpos - 1 );

    while ( !backtrackstop[outputpos] ) {
      // possibilities for this atom have been exhausted, backtrack to
      // previous backtracking position.
 //      if ( outputpos > 1 )
//        determineBacktrackPos ( node, (*ftd->changes.assignments) [ outputpos - 1 ] );
      //minpos--;
      outputpos--;
      if ( outputpos <= 0 ) {
        undoAssignments ();
        return;
      }
      node = nodeHistory.outputNode ( outputpos - 1 );
      removeFromNrVariables  ( node );
      el = element ( outputpos - 1 );
    }
  }
  else {
    undoAssignments ();
    return;
  }

  el = el->next;

  while ( true ) {
    el = findFirstValidAssignment ( node, el );

    if ( el ) {
      // assign variables.
      ftd->applyToAssignments ( node, el );
      set ( outputpos - 1, el );
      if ( evaluate_nooutputs ( outputpos - 1 ) ) {
	// goto next position (if available)
	// furthermore take into account the variables
	// that have been assigned now.
	if ( outputpos == startpos ) {
	  addToNrVariables ( node );
	  if ( !evaluateExpansions () ) {
	    undoAssignments (); // we can stop
	    return;
	  }
	  removeFromNrVariables ( node );

	  el = el->next;
	  continue;
	}
	else {
	  backtrackstop[outputpos] = false;
	  outputpos++;
	  addToNrVariables ( node );
	  node = nodeHistory.outputNode ( outputpos - 1 );
	  el = evaluateOutputs ( node );
	  if ( el )
	    continue; // do not backtrack in the next loop.
	}
      }
      else {
        if ( backtrackstop[outputpos] ) {
	  el = el->next;
	  continue;
	}
      }
    }
    backtrackPosOnInputs ( node );
      if ( outputpos <= minpos ) {
        if ( outputpos > 1 )
          determineBacktrackPos ( node, NULL ); // (*ftd->changes.assignments) [ outputpos - 1 ] );
	minpos = outputpos;
      }
    do {
      outputpos--;
      if ( outputpos <= 0 ) {
        undoAssignments ();
        return;
      }
      node = nodeHistory.outputNode ( outputpos - 1 );
      removeFromNrVariables  ( node );
      el = element ( outputpos - 1 );
    }
    while ( !backtrackstop[outputpos] );
    el = el->next;
  }
}


void applyAllAssignments () {
  TidList *tidundos = new TidList;
  Node *node = nodeHistory.last().node;
  TidList *tidlist = node->tidlist;
  FListWalk<TidData> w ( &tidlist->tids );
  TidData *tidd, *undotidd;
  nodeHistory.assignundos.append ( tidundos );
  int back = modetable.nrOutput ( node->mode ) > 0 ? -1 : 0;

  while ( tidd = w.current () ) {
    undotidd = fulltidarray->tids[tidd->tid]->undoApplyData ( *tidd, back, 0 );
    fulltidarray->tids[tidd->tid]->apply ( *tidd, back, 0 );
    tidundos->tids.append ( *undotidd );
    undotidd->assignments = NULL; // prevent deletion.
    w.next ();
  }
}

void undoAllAssignments () {
  TidList *tidundos = nodeHistory.assignundos[nodeHistory.assignundos.size () - 1];
  FListWalk<TidData> w ( &tidundos->tids );
  TidData *undotidd;
  Node *node = nodeHistory.last().node;
  int back = modetable.nrOutput ( node->mode ) > 0 ? 1 : 0;

  while ( undotidd = w.current () ) {
    fulltidarray->tids[undotidd->tid]->apply ( *undotidd, back, 0 );
    w.next ();
  }
  
  delete tidundos;
  nodeHistory.assignundos.releaselast ();
}

bool frequentTidUnion ( TidList *list1, TidList *list2 ) {
  FListWalk<TidData> w1 ( &list1->tids ), w2 ( &list2->tids );
  int size1 = list1->tids.size (), size2 = list2->tids.size ();
  int newsize = 0;
  TidData *tidd1 = w1.current (), *tidd2 = w2.current ();
  
  while ( newsize < minsup && newsize + min ( size1, size2 ) >= minsup && 
          tidd1 && tidd2 ) {
    if ( tidd1->tid < tidd2->tid ) {
      w1.next ();
      tidd1 = w1.current ();
      size1--;
    }
    else
      if ( tidd1->tid > tidd2->tid ) {
        w2.next ();
	tidd2 = w2.current ();
	size2--;
      }
      else {
        w1.next ();
	tidd1 = w1.current ();
	w2.next ();
	tidd2 = w2.current ();
	size1--;
	size2--;
	newsize++;
      }
  }
  return newsize >= minsup;
}
